;################################################################## RCVHEX.ASL
; Purpose.....: Receive data in intel-HEX format over USART on EOM-MPF and
;		store at addresses defined in HEX file.
;		Serial parameters are 1200,N,8,1. With this baudrate it works
;		on systems which do not support RTS handshake.
; Target......: MPF-1P
; Assembler...: ASL V1.42 Beta
; Command Line:	ASL -listline-prefix %04a -L -U -olist <name>.prn <name>.asl
; PsPadParser.: > > > %F(%L):%C: error:
;#############################################################################

;********************************************************************* history
;=========================================================== V01.00:27.02.2024
; Created
;=========================================================== V01.10:03.04.2014
; Added sign-on message
;=========================================================== V01.20:06.04.2024
; Added abort with shift and control keys
;=========================================================== V01.30:07.04.2024
; Use lKSAC() instead of local function
;=========================================================== V01.40:08.04.2024
; Moved variables to MPF RAM to get continuous RAM from D800 to FD00
;=========================================================== V01.40:27.08.2024
; Changed to ASL assembler, no code change
;***************************************************************** end history

;***************************************************************** definitions
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++ assembler directives
		CPU	Z80		; use Z80 CPU
		PAGE	0		; no formfeeds in listing file
		
;+++++++++++++++++++++++++++++++++++ hardware, monitor and library definitions
		LISTING	OFF
		INCLUDE	../LIBRARY/MPF1P.DEF
		LISTING ON
		
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ build addresses
wRHCS:		EQU	0B160H		; code (B160-B325 if IOM-ROM)
wRHDS:		EQU	0FD00H		; data on top of MPF RAM

;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ misc
bRHSTZ:		EQU	020H		; application stack size
bRHRBZ:		EQU	80		; receive line buffer size

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ variables
vRHDP:		EQU	wRHDS 		; display progress pattern
vRHFA:		EQU	vRHDP + 1	; from address for success display
vRHBC:		EQU	vRHFA + 2	; receive byte counter
vRHRB:		EQU	vRHBC + 2	; receive line buffer
wRHVE:		EQU	vRHRB + bRHRBZ	; variables end address

;======================================== check overlap with library variables
		IF	wRHVE > dwLVST && MomPass = 2
		WARNING	"Variables overlap with library:\{wRHVE} >> \{dwLVST}"
		ENDIF
;************************************************************* end definitions

;************************************************************** implementation
		ASEG
		ORG	wRHCS
		
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ main program
;======================================================================== RH()
; Purpose.....: Main program
; Labels......: R)eceive H)ex
;=============================================================================
RH:		CALL	dfLLS		; initialize library variables

		LD	HL,0
		LD	(vRHFA),HL	; clear start address
		LD	(vRHBC),HL	; clear received byte counter	

		LD	IX,tRHSO	; sign-on message
		LD	B,030H		; 30H ~1s
		CALL	dfLDT		; show it

;---------------------------------------------------------------- wait pattern
		LD	A,'.'
		LD	(vRHDP),A	; save for pattern routine
		CALL	RHDP		; display wait pattern

;----------------------------------------------------------------- setup USART
		LD	B,dbUB1K	; baudrate 1200 (works without HS)
		LD	C,dbU8N1	; 8 data, no parity, 1 stop	
		CALL	dfLUS		; setup USART

;------------------------------------------------------ read a line from USART
RHRL:		SCF			; set RTS
		CALL	dfLURTS		; active

		LD	HL,vRHRB	; HL= buffer address
		LD	E,bRHRBZ	; E= buffer size

RHRLL:		CALL	dfLKSAC		; check for abort request
		JP	C,RHQ		; shift and control pressed, exit

		CALL	dfLURB		; get byte from USART
		JP	C,RHER		; CY=1: some error
		JR	Z,RHRLL		; Z=1: no character ready

		CP	0AH		; end of line?
		JR	Z,RHPL		; yes, process read line
		LD	(HL),A		; no, store in buffer
		INC	HL		; bump buffer pointer
		DEC	E		; dec character counter
		JP	Z,RHEB		; buffer is full, no LF seen
		JR	RHRLL		; not full, wait for next

;----------------------------------------------------------- process read line
RHPL:		XOR	A		; reset
		CALL	dfLURTS		; RTS line

		CALL	RHDP		; update progress display

		LD	DE,vRHRB	; DE= buffer start
		LD	L,bRHRBZ	; L= buffer size
			
;------------------------------------ search for and skip line start character
RHPS:		LD	A,(DE)		; A= next character
		INC	DE
		CP	':'		; start of line?
		JR	Z,RHPN		; yes, process line
		DEC	L		; whole buffer searched?
		JR	NZ,RHPS		; no, go for next
		JR	RHEN		; yes, error, no colon found

;--------------------------------------------- extract number of bytes in line
RHPN:		LD	C,0		; C= init checksum counter

		LD	A,2		; convert 2 digits
		CALL	RHH2B		; convert HEX to bin
		LD	B,L		; B= number of bytes in this line
		XOR	A		; end record
		OR	B		; has null bytes in line
		JR	Z,RHS		; it is, success, exit

;--------------------------------------------- maintain received bytes counter
		PUSH	DE		; save receive line pointer
		EX	DE,HL		; (D)E= number of bytes in line
		LD	D,0		; must be 0 for ADD
		LD	HL,(vRHBC)	; HL= byte counter
		ADD	HL,DE		; add this line
		LD	(vRHBC),HL	; and write back
		POP	DE		; recall receive line pointer
	
;------------------------------------------------------- extract store address
		LD	A,4		; convert 4 digits
		CALL	RHH2B		; convert HEX to bin

		CALL	RHCKM		; system RAM overwrite?
		JR	C,RHEY		; yes, exit with error

		PUSH	HL
		POP	IX		; IX= store address

;------------------------------------- store start address for success display
		LD	A,(vRHFA+1)	; A= hibyte start address
		OR	A		; already stored?
		JR	NZ,RHT		; yes
		LD	(vRHFA),HL	; no, store it

;--------------------------------------------------------- extract record type
RHT:		LD	A,2		; convert 2 digits
		CALL	RHH2B		; convert HEX to bin
		LD	A,L
		CP	01H		; end record?
		JR	Z,RHRL		; yes, read end record

;--------------------------------------------------- read and store data bytes
RHPDL:		LD	A,2		; convert two digits
		CALL	RHH2B		; convert HEX to bin
		LD	(IX+0),L	; save to memory
		LD	A,(IX+0)	; compare memory
		CP	L		; with saved
		JR	NZ,RHEV		; not same, no RAM or ROM
		INC	IX		; point to next address
		DJNZ	RHPDL		; go for next

;----------------------------------------------------------- validate checksum
		LD	A,2		; convert checksum digits
		CALL	RHH2B		; get received checksum
		XOR	A
		OR	C		; sum of all binary bytes must be 0
		JR	NZ,RHES		; not zero, error, exit

		JP	RHRL		; go for next line

;-------------------------------------- success, display from and to addresses
RHS:		CALL	dfMCLR		; clear display buffer
		LD	HL,tRHFP	; from prompt
		CALL	dfMMSG		; to display buffer
		LD	HL,(vRHFA)	; HL= from address
		CALL	dfMB4H		; to display buffer
		LD	HL,tRHTP	; to prompt
		CALL	dfMMSG		; to display buffer
		PUSH	IX
		POP	HL		; HL= last written address
		CALL	dfMB4H		; to display buffer
		JR	RHX		; and display, wait for key

;--------------------------------------------------------------- error handler
RHEN:		LD	DE,tRHEN	; no colon (start mark) found
		JR	RHMSG
RHES:		LD	DE,tRHES	; checksum error
		JR	RHMSG
RHEV:		LD	DE,tRHEV	; verify error
		JR	RHMSG
RHEY:		LD	DE,tRHEY	; sys RAM overwrite attempt
		JR	RHMSG
RHEB:		LD	DE,tRHEB	; receive buffer full
		JR	RHMSG
RHER:		LD	DE,tRHER	; USART error

RHMSG:		CALL	dfMCLR		; clear display buffer
		PUSH	IX
		POP	HL		; HL= last written address
		CALL	dfMB4H		; last written address to display buf		
		EX	DE,HL		; HL= error message
  		CALL	dfMMSG		; error message to display buffer

;---------------------------------- display exit message and return to monitor
RHX:		LD	IX,dvMDB	; IX= display buffer address
		CALL	dfMSC		; print message and wait for key

RHQ:		XOR	A
		CALL	dfLURTS		; disable handshake
		JP	dwMRET		; monitor warm start

;-------------------------------------------------------------- text constants
tRHSO:		DB	'    RH V24-08-27    '
tRHFP:		DB	'FROM ',           0DH
tRHTP:		DB	'  TO ',           0DH
tRHEN:		DB	' NO START MARK',  0DH
tRHEB:		DB	' LINE TOO LONG',  0DH 
tRHES:		DB	' CHECKSUM ERROR', 0DH
tRHEV:		DB	' VERIFY ERROR',   0DH
tRHEY:		DB	' SYS-RAM OVR',    0DH
tRHER:		DB	' USART ERROR',    0DH
;==================================================================== eoM RH()
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ end main program

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ subroutines
;===================================================================== RHCKM()
; Purpose.....: Check if system RAM is going to be overwitten
; Input.......:	HL: current write address
;		B:  number of bytes to write (one HEX line)
; Output......:	CY: 0= no overwite
;		    1= overwrite will occur
; Registers...:	AF destroyed
; Labels......: R)ead H)ex C)ecK) M)emory
;=============================================================================
RHCKM:		PUSH	HL
		PUSH	BC

		LD	C,B
		LD	B,0		; BC= bytes going to be written
		ADD	HL,BC		; HL= last address to write
		LD	BC,dwMSTB	; BC= user stack bottom address
		SCF			; set CY for SBC
		SBC	HL,BC		; CY=1 if BC > HL
		CCF			; CY=1 if BC <= HL

		POP	BC
		POP	HL
		RET
;================================================================= eoF RHCKM()

;===================================================================== RHH2B()
; Purpose.....:	Convert HEX digits to binary using monitor function
; Input.......:	A:  Number of digits to convert
;		C:  current checksum counter
;		DE: Points to 1st digit
; Output......:	HL: Binary number
;		DE: Incremented by number of digits
;		C:  The number of HEX digits added 
; Registers...:	AF, DE, C, HL destroyed
; Labels......: R)ead H)ex H)ex 2)to B)in
;=============================================================================
RHH2B:		PUSH	BC

;------------------------------------------------ insert terminator for HEXBIN
		LD	H,0
		LD	L,A		; L= number of digits
		ADD	HL,DE		; HL= terminator position
		LD	C,(HL)		; save byte at this position
		LD	(HL),0		; set terminator for HEXBIN
		PUSH	HL		; save for returning in DE

;---------------------------------- convert to binary using monitor's HEXBIN()
		PUSH	BC
		PUSH	HL		; save position for restore
		CALL	dfMH2B		; call monitor's HEXBIN, HL=binary
		POP	DE		; DE= terminator position
		POP	BC

		EX	DE,HL		; HL= terminator position
		LD	(HL),C		; restore original byte
		EX	DE,HL		; HL= converted binary
		POP	DE		; DE= incremented buffer pointer

;---------------------------------------------------------- calculate checksum
		POP	BC		; recall checksum
		LD	A,C		; A= checksum
		ADD	A,H		; add hibyte of result
		ADD	A,L		; add lobyte of result
		LD	C,A		; back to checksum counter

		RET
;================================================================= eoF RHH2B()

;====================================================================== RHDP()
; Purpose.....:	Display static pattern in the center of the display to
;		signal progress. On every call the pattern changes from
;		'(' to ')' and vice versa
; Registers...:	AF destroyed
; Labels......: R)ead H)ex D)isplay P)attern
;=============================================================================
RHDP:		PUSH	DE
		PUSH	BC

RHDPC:		LD	D,00011111B	; digits 8-16
		LD	E,10000000B	; digits 1-8
		LD	C,00000000B	; digits 17-20
		LD	A,(vRHDP)	; get pattern to display
		LD	B,A		; save pattern
		CALL	dfLDP		; display the pattern

		LD	A,'('
		CP	B		; is '(' displayed?
		JR	NZ,RHDPS	; no, display on next call
		LD	A,')'		; yes, display ')' on next call
RHDPS:		LD	(vRHDP),A	; save for next call
		XOR	A		; normal exit, reset CY

RHDPX:		POP	BC
		POP	DE
		RET
;================================================================== eoF RHDP()
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ end subroutines

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ size info
wRHCSE:		EQU	$

		IF	MomPass = 2
		MESSAGE "Code from \{wRHCS} to \{wRHCSE} Size \{wRHCSE-wRHCS}"
		MESSAGE "Data from \{wRHDS} to \{wRHVE} Size \{wRHVE-wRHDS}"
		MESSAGE "     remaining \{dwLVST-wRHVE}"
		ENDIF
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ end size info

;############################################################## EOF RCVHEX.ASL
		END
